import re

class Node:
    def __init__(self, indentation, line = None):
        self.attributes   = {}
        self.children     = []
        self.indentation  = indentation
        self.last_result  = None
        self.line         = line
        self.parts        = []
        self.parent       = None
        self.prev_sibling = None
        self.post_sibling = None
        self.tag          = ""
        self.text         = ""
        self.ws_prepend   = False
        self.ws_append    = False

        if line is None:
            return

        self.line  = line.strip()
        self.parts = self.line.split(" ")

        if self.line == '':
            return

        self.parse_tag()
        self.parse_attributes()

    def parse_tag(self):
        if len(self.parts) == 0:
            breakpoint()

        tag_part = self.parts[0]

        if tag_part == 'javascript':
            self.tag = 'script'
            self.attributes['type'] = 'text/javascript'
            return

        if tag_part[0] == '-':
            return self.parse_logic_tag()

        if tag_part[-1] == '>':
            self.ws_append  = True
            tag_part        = tag_part[:-1]

        if tag_part[-1] == '<':
            self.ws_prepend = True
            tag_part        = tag_part[:-1]

        if tag_part[0] == '/':
            self.tag = 'comment'
            return

        classes = []
        e = re.split(r'([.#])', tag_part)

        self.tag = e[0]

        for i in range(len(e)):
            if e[i] == '#':
                self.attributes["id"] = e[i + 1]
            if e[i] == '.':
                classes.append(e[i + 1])

        if len(classes) > 0:
            class_names = " ".join(classes)
            if self.attributes.get("class"):
                self.attributes["class"] += f" {class_names}"
            else:
                self.attributes["class"] = class_names

    def parse_logic_tag(self):
        match self.parts[1]:
            case 'if':
                return self.parse_if_tag()
            case 'else':
                return self.parse_else_tag()
            case 'for':
                return self.parse_for_tag()
            case "render":
                return self.parse_render_tag()

    def parse_render_tag(self):
        self.tag = 'render'
        self.attributes['template'] = self.parts[2]
        return

    def parse_if_tag(self):
            self.tag = 'if'
            self.attributes['condition'] = " ".join(self.parts[2:])
            return

    def parse_else_tag(self):
            self.tag = 'else'
            return

    def parse_for_tag(self):
            self.tag                 = 'for'
            self.attributes['var']   = self.parts[2]
            self.attributes['array'] = self.parts[4]
            return

    def parse_attributes(self):
        if self.parts[0] == '-':
            return

        for a in (self.parts[1:]):
            if '=' in a:
                key, val = a.split('=')
                self.attributes[key] = val.strip('"').strip("'")
            else:
                self.text += a + " "

        self.text = self.text.strip()

    def add_node(self, node):
        node.parent = self

        self.children.append(node)

        return node

    def add_line(self, line):
        self.children.append(Node(self.indentation + 1, line))

    def tag(self):
        return self.tag

    def has_attributes(self):
        return self.attributes != []

    def attributes(self):
        return ' '.join(self.attributes)



from .node import Node
from .node_parser import NodeParser
from collections import deque
from bs4 import BeautifulSoup
import os
import pprint

class TrimTemplate:
    def __init__(self, template, pretty=True, debug='all', indent=4, vars={}):
        self.dir        = ""
        self.nodes      = []
        self.parsed     = ""
        self.variables  = vars
        self.options    = {
            "debug":  debug,
            "pretty": pretty,
            "indent": indent
        }
        # template can be a string or a path to a file
        if os.path.isfile(template):
            self.read_template_file(template)
            self.dir = os.path.dirname(template)
        else:
            self.lines = template.splitlines()

        self.parse_lines()

    def clone(self, template):
        return TrimTemplate(template, vars = self.variables)

    def set(self, key, value = None):
        if isinstance(key, dict):
            self.variables.update(key)
        else:
            self.variables[key] = value

    def read_template_file(self, path):
        with open(path, "r") as text:
            self.lines = text.readlines()

    def parse_lines(self):
        node = self.node = Node(-1)
        node.parent = node
        node.tag = "root"

        lines = deque(self.lines)

        while len(lines) > 0:
            line = lines.popleft()
            stripped = line.strip()

            if stripped == "":
                continue

            if stripped[0] == "/":
                if stripped[1] == "!":
                    node.add_node(Node(node.indentation, line))
                continue

            indentation = len(line) - len(line.lstrip())
            new_node    = Node(indentation, line)

            if indentation > node.indentation:
                node = node.add_node(new_node)
            elif indentation == node.indentation:
                node = node.parent.add_node(new_node)
            else:

                while indentation <= node.indentation:
                    node = node.parent

                new_node.prev_sibling = node
                node = node.parent.add_node(new_node)

            # if the node is an embedded script, read the remaining lines
            if node.tag in ['css:', 'javascript:']:
                # breakpoint()
                while (len(lines[0]) - len(lines[0].lstrip())) > node.indentation:
                    node.text += (lines.popleft()) + "\n"
                print(node.__dict__)

        self.nodes.append(node)


    def parse_nodes(self):
        self.parsed += NodeParser(self.node, self).parse()

    def render(self):
        self.parse_nodes()

        if self.options["pretty"]:
            return BeautifulSoup(self.parsed, 'html.parser').prettify()

        return self.parsed

    def left_indent(self, text, indentation = 0):
        if text == "" or text is None:
            return ""
        return ''.join([(indentation * ' ') + line for line in text.splitlines(True)])

    def debug_node(self, node):
        if node.indentation < 0:
            print('root node')
        else:
            pp = pprint.PrettyPrinter(indent = node.indentation + 1)
            if self.options['debug'] == 'all':
                pp.pprint(node.__dict__)
            else:
                print(' ' * (node.indentation + 4) + node.tag)

        for node in node.children:
            self.debug_node(node)

    def debug(self):
        print(self.__dict__)
        self.debug_node(self.node)


# from trim import TrimTemplate
import os

SINGLE_TAGS = ["doctype", "img", "br", "hr", "input", "link", "meta"]
BOOLEAN_ATTRIBUTES = [
    'checked', 'selected', 'disabled', 'readonly', 'multiple', 'ismap', 'defer', 'declare', 'noresize', 'nowrap',
]

DOCTYPES = {
    "html":         '<!DOCTYPE html>',
    "basic":        '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML Basic 1.1//EN" "http://www.w3.org/TR/xhtml-basic/xhtml-basic11.dtd">',
    "frameset":     '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">',
    "mobile":       '<!DOCTYPE html PUBLIC "-//WAPFORUM//DTD XHTML Mobile 1.2//EN" "http://www.openmobilealliance.org/tech/DTD/xhtml-mobile12.dtd">',
    "strict":       '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',
    "transitional": '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',
}


class NodeParser:
    def __init__(self, node, engine=None):
        self.node      = node
        self.parsed    = ""
        self.engine    = engine
        self.variables = engine.variables.copy()

    def parse(self):
        match self.node.tag:
            case 'comment':
                return f"<!-- {self.node.text} -->"
            case 'css:':
                return f"\n<style>\n{self.node.text}</style>"
            case "doctype":
                return DOCTYPES.get(self.node.text, "")
            case "else":
                return self.parse_else()
            case "javascript:":
                return f"\n<script type='javascript'>\n{self.node.text}\n</script>"
            case 'javascript:':
                return f"\n<script type='javascript'>\n{self.node.text}</script>"
            case 'if':
                return self.parse_if()
            case "for":
                return self.parse_for()
            case "render":
                dir      = self.engine.dir
                template = dir + '/' + self.node.attributes.get("template")
                renderer = self.engine.clone(template)
                # breakpoint()
                return renderer.render()
            case "root":
                return self.parse_children()

        self.parse_node()

        return self.parsed

    def parse_node(self):
        self.parsed += f"<{self.node.tag}"
        self.parsed += self.parse_attributes()

        if self.node.tag in SINGLE_TAGS:
            self.parsed += self.close_node()
        else:
            self.parsed += ">"
            self.parse_children()

        self.parsed += self.parse_text()

        if self.node.tag not in SINGLE_TAGS:
            self.parsed += self.close_node()

        return self.parsed

    def parse_for(self):
        var   = self.node.attributes.get("var", "")
        array = self.node.attributes.get("array", "")

        for t in self.engine.variables.get(array):
            self.variables[var] = t
            for node in self.node.children:
                self.parsed += NodeParser(node, self).parse()

        return self.parsed

    def parse_if(self):
        condition = self.node.attributes.get("condition", "False")
        result    = eval(condition, self.engine.variables)

        if result:
            return self.parse_children()

        self.node.last_result = result

        return ""

    def parse_else(self):
        if not self.node.prev_sibling.last_result:
            return self.parse_children()

        return ""

    def parse_text(self):
        t = self.node.text

        if t == "":
            return ""

        if self.node.ws_prepend:
            t = " " + t

        if "{" in t:
            t = t.format(**self.engine.variables)

        if self.node.ws_append:
            t += " "

        return t

    def parse_children(self):
        for node in self.node.children:
            self.parsed += NodeParser(node, self.engine).parse()

        return self.parsed

    def close_node(self):
        tag = self.node.tag
        if tag == "":
            return ""
        if tag in SINGLE_TAGS:
            return "/>"
        else:
            return f"</{tag}>"

    def parse_attributes(self):
        p = ""
        if self.node.has_attributes():
            for a, v in self.node.attributes.items():
                if v == "":
                    continue

                if "{" in v:
                    v = v.format(**self.engine.variables)

                if a in BOOLEAN_ATTRIBUTES:
                    v = self.boolean_attribute(a, v)
                    if v is False:
                        continue

                p += f' {a}="{v}"'

        return p

    def boolean_attribute(self, attribute, value):
        if attribute in BOOLEAN_ATTRIBUTES:
            if value == "True":
                return attribute

        if self.engine.variables.get(value, False):
            return attribute

        return False
